/* 
 * Copyright (c) 2001, 2002, 2003, 2004 James Bailie <jimmy@jamesbailie.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote 
 * products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "lisp.h"
#include "intrinsics.c"

void sigchld_handler( int signo )
{
   int pid;

   while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      if ( pid == lisp_pid )
         lisp_pid = -1;
}

void sigterm_handler( int signo )
{
   exit( 0 );
}

void sigint_handler( int signo )
{
   if ( !blocked )
      exit( 0 );
}

void print_err( enum err_types err, char *func, int arg, enum err_types type)
{
   fprintf( stderr, error_messages[ err - 11 ], func, arg, 
            ( type < 0 ? "" : err_strings[ type ] ));
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      fprintf( stderr, "memory: malloc: %s.\n", strerror( errno ));
      exit( 1 );
   }

   return ptr;
}

char *str_dup( char *str, int len )
{
   char *ptr;

   ptr = ( char *)memory( len + 1 );
   bcopy( str, ptr, len );
   ptr[ len ] = '\0';

   return ptr;
}

struct object *make_object()
{
   if ( reclaimed_objects->used )
      return stack_pop( reclaimed_objects );

   if ( object_pool_free == 0 )
   {
      stack_push( object_pool_stack, ( void *)object_pool );
      object_pool = ( struct object *)memory( sizeof( struct object ) * POOL_INC );
      object_pool_ptr = object_pool;
      object_pool_free = POOL_INC;
   }

   --object_pool_free;

   object_pool_ptr->data.head = object_pool_ptr->next = NULL;
   object_pool_ptr->flags = 0;

   return object_pool_ptr++;
}

struct tree *make_atom()
{
   if ( reclaimed_atoms->used )
      return ( struct tree *)stack_pop( reclaimed_atoms );

   if ( atom_pool_free == 0 )
   {
      stack_push( atom_pool_stack, ( void *)atom_pool );
      atom_pool = ( struct tree *)memory( sizeof( struct tree ) * POOL_INC );
      atom_pool_ptr = atom_pool;
      atom_pool_free = POOL_INC;
   }

   --atom_pool_free;

   atom_pool_ptr->syntax = NULL;
   atom_pool_ptr->len = 0;
   atom_pool_ptr->id = 0;
   atom_pool_ptr->flags = 0;
   atom_pool_ptr->data.hash = NULL;
   atom_pool_ptr->left = atom_pool_ptr->right = NULL;

   return atom_pool_ptr++;
}

struct stack *get_hash_keys( struct hash_elt **table )
{
   struct stack *stk;
   struct hash_elt *ptr;
   int i;

   stk = make_stack();

   for( i = 0; i < HASH_SIZE; ++i )
   {
      ptr = table[ i ];

      while( ptr != NULL )
      {
         stack_push( stk, ( struct object *)ptr->binding );
         ptr = ptr->next;
      }
   }

   return stk;
}

struct stack *get_hash_values( struct hash_elt **table )
{
   struct stack *stk;
   struct hash_elt *ptr;
   int i;

   stk = make_stack();

   for( i = 0; i < HASH_SIZE; ++i )
   {
      ptr = table[ i ];

      while( ptr != NULL )
      {
         stack_push( stk, ( struct object *)ptr->element );
         ptr = ptr->next;
      }
   }

   return stk;
}

void insert_elt( struct hash_elt **hash, int id, struct object *elt )
{
   int key;
   struct hash_elt *ptr = NULL, *ptr2 = NULL;

   key = ( id + HASH_SIZE ) % HASH_SIZE;

   if ( hash[ key ] == NULL )
   {
      hash[ key ] = memory( sizeof( struct hash_elt ));
      hash[ key ]->element = elt;
      hash[ key ]->next = NULL;
      hash[ key ]->binding = id;
      return;
   }

   ptr2 = hash[ key ];

   for( ptr = ptr2; ptr != NULL; ptr = ptr->next )
      if ( ptr->binding == id )
         break;
      else
         ptr2 = ptr;

   if ( ptr == NULL )
   {
      ptr2->next = ( struct hash_elt *)memory( sizeof( struct hash_elt ));
      ptr = ptr2->next;
      ptr->next = NULL;
   }

   ptr->element = elt;
   ptr->binding = id;
}

void *lookup_elt( struct hash_elt **hash, int id )
{
   int key;
   struct hash_elt *ptr;

   key = ( id + HASH_SIZE ) % HASH_SIZE;

   if ( hash[ key ] == NULL )
      return NULL;

   ptr = hash[ key ];

   do
   {
      if ( ptr->binding == id )
         return ptr->element;
      
      ptr = ptr->next;
   }
   while( ptr != NULL );

   return NULL;
}

void remove_elt( struct hash_elt **hash, int id )
{
   int key;
   struct hash_elt *ptr, *ptr2;

   key = ( id + HASH_SIZE ) % HASH_SIZE;

   if ( hash[ key ] == NULL )
      return;

   ptr = hash[ key ];
   ptr2 = NULL;

   do
   {
      if ( ptr->binding == id )
         break;

      ptr2 = ptr;
      ptr = ptr->next;
   }
   while( ptr != NULL );

   if ( ptr == NULL )
      return;

   if ( ptr2 == NULL )
      hash[ key ] = ptr->next;
   else 
      ptr2->next = ptr->next;

   free( ptr );
}

struct tree *get_id( char *name, int len )
{
   struct tree *ptr;
   int test;

   if ( syntax_tree == NULL )
   {
      syntax_tree = make_atom();
      syntax_tree->syntax = str_dup( name, len );
      syntax_tree->len = len;
      syntax_tree->id = atom_counter++;
   
      insert_atom( syntax_tree->id, syntax_tree );

      return syntax_tree;
   }

#ifdef DEBUG
   assert( *syntax_tree->syntax != '\0' );
#endif

   ptr = syntax_tree;

AGAIN:
   test = memcmp( name, ptr->syntax, MIN( ptr->len, len ) + 1 );

   if ( test == 0 )
      return ptr;

   if ( test > 0 )
   {
      if ( ptr->right != NULL )
      {
         ptr = ptr->right;
         goto AGAIN;
      }

      ptr->right = make_atom();
      ptr = ptr->right;
   }
   else 
   {
      if ( ptr->left != NULL )
      {
         ptr = ptr->left;
         goto AGAIN;
      }

      ptr->left = make_atom();
      ptr = ptr->left;
   }

   ptr->syntax = str_dup( name, len );
   ptr->len = len;
   ptr->id = atom_counter++;

   insert_atom( ptr->id, ptr );

   return ptr;
}

void hash_free( struct hash_elt **hash )
{
   int i;
   struct hash_elt **ptr, *ptr2, *ptr3;

   ptr = hash;

   for( i = 0; i < HASH_SIZE; ++i )
   {
      if ( *ptr == NULL )
      {
         ++ptr;
         continue;
      }

      ptr2 = *ptr;

      do
      {
         ptr3 = ptr2->next;
         free( ptr2 );
         ptr2 = ptr3;
      }
      while( ptr2 != NULL );

      ++ptr;
   }
}

void free_syntax( struct tree *tree )
{
   if ( tree != NULL )
   {
      free( tree->syntax );

      switch( type( tree->flags ))
      {
         case ATOM_STRING:
            free( tree->data.string->string );
            free( tree->data.string );
            break;

         case ATOM_REGEXP:
            regfree( tree->data.regexp );
            free( tree->data.regexp );
            break;

         case ATOM_TABLE:
            hash_free( tree->data.hash );
            free( tree->data.hash );
            break;

         case ATOM_DB:
#ifdef SQL
            if ( tree->data.db != NULL )
               sqlite3_close( tree->data.db );
#endif
            break;

         case ATOM_STACK:
            stack_free( tree->data.stack );
            break;

         case ATOM_NUMBER:
            free( tree->data.value );
            break;

         case ATOM_CLOSURE:
         case ATOM_MACRO:
            free( tree->data.closure );
            break;

         case ATOM_ACT_RECORD:
            stack_free( tree->data.act_record );
      }

      free_syntax( tree->left );
      free_syntax( tree->right );
   }
}

struct stack *make_stack()
{
   struct stack *a;

   a = ( struct stack *)memory( sizeof( struct stack ));
   a->values = memory( sizeof( void * ) * stack_inc );
   a->free = stack_inc;
   a->used = 0;
   a->top = a->values;

   return a;
}

void string_free( struct string *s )
{
   free( s->str );
   free( s );
}

struct string *make_string()
{
   struct string *s;

   s = ( struct string *)memory( sizeof( struct string ));
   s->str = ( char *)memory( stack_inc );
   *s->str = '\0';
   s->free = stack_inc - 1;
   s->used = 0;
   s->top = s->str;

   return s;
}

int load( char *filename )
{
   int count;
   int file;

AGAIN:
   file = open( filename, O_RDONLY | O_NONBLOCK | O_SHLOCK );

   if ( file == -1 )
   {
      if ( errno == EAGAIN )
      {
         sleep( 1 );
         goto AGAIN;
      }

      fprintf( stderr, "load: %s: %s\n", filename, strerror( errno ));
      return 1;
   }

   for( count = 1; ; ++count )
   {
      int depth;

      depth = parse( file );

      if ( depth > 0 )
         break;
      else if ( depth < 0 )
         fprintf( stderr, "%d extra ')'\n", -depth );

      if ( evaluate() )
      {
         if ( !stop )
            fprintf( stderr, "load: evaluation of expression %d in %s "
                             "failed.\n", count, filename );

         close( file );
         get_token( 0, -1 );
         return 1;
      }

      stack_pop( stack );
   }

   close( file );
   stack_push( stack, make_atom_from_number( 1 ));

   return 0;
}

void init_intrinsics()
{
   int ( **fp )( char *, struct object *);
   char **sp;
   struct tree *entry1, *entry2;
   struct object *object;

   sp = intrinsic_syntax;

   for( fp = intrinsics; *fp != NULL; ++fp )
   {
      entry1 = get_id( *sp, strlen( *sp ));
      entry1->flags = ATOM_INTRINSIC;
      set( entry1->flags );
      entry1->data.function = *fp;

      object = make_object();
      mark( object->flags );
      object->data.atom = entry1;

      ++sp;

      entry2= get_id( *sp, strlen( *sp ));
      entry2->flags = ATOM_SYMBOL;
      set( entry2->flags );
      entry2->data.value = ( double *)1;

      insert_binding( entry2->id, object );

      ++sp;
   }
}

void initialize( int argc, char **argv )
{
   int i, result;
   struct hash_elt **ptr;
   struct tree *tree;
   struct string *s;
   char err[ 80 ], *ptr2;
#ifdef TERM
   struct sigaction sigact;
#endif
   struct stat stats;

   signal( SIGPIPE, SIG_IGN );
   signal( SIGTERM, sigterm_handler );
   signal( SIGINT, sigint_handler );
   signal (SIGCHLD, sigchld_handler );

#ifdef TERM
   sigact.sa_handler = sigwinch_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGWINCH, &sigact, NULL ) < 0 )
      fprintf( stderr, "initialize: sigaction: %s.\n", strerror( errno ));

   if ( isatty( STDOUT_FILENO ) && 
        setupterm( NULL, 1, &result ) == OK )
   {
      cl = tigetstr( "clear" );
      ce = tigetstr( "el" );
      cm = tigetstr( "cup" );
      vi = tigetstr( "civis" );
      ve = tigetstr( "cnorm" );
      sf = tigetstr( "ind" );
      sr = tigetstr( "ri" );
      al = tigetstr( "il1" );
      sc = tigetstr( "sc" );
      rc = tigetstr( "rc" );
   }
#endif

   if ( isatty( STDOUT_FILENO ) && 
        tcgetattr( STDOUT_FILENO, &canon_termios ) < 0 )
   {
      fprintf( stderr, "initialize: tcgetattr: %s.\n", strerror( errno ));
      return;
   }

   srandomdev();

   first_arg = arg_ptr = argv;
   last_arg = &argv[ argc - 1 ];

   result = regcomp( &merge_regex, "[^/]+$", REG_EXTENDED );

   if ( result )
   {
      regerror( result, &merge_regex, err, sizeof( err ));
      fprintf( stderr, "fatal error: initialize: merge_regex: regcomp(): %s.\n",
               err );
      exit( 1 );
   }

   result = regcomp( &find_poss_regex, 
                     "^(/?([^/]+/)*)?([^/]+)?", REG_EXTENDED );

   if ( result )
   {
      regerror( result, &find_poss_regex, err, sizeof( err ));
      fprintf( stderr, "fatal error: initialize: regcomp(): %s.\n", err );
      exit( 1 );
   }

   ptr = atoms;

   for( i = 0; i < HASH_SIZE; ++i )
      *ptr++ = NULL;

   ptr = env;

   for( i = 0; i < HASH_SIZE; ++i )
      *ptr++ = NULL;

   syntax_tree = NULL;

   reclaimed_objects = ( struct stack *)memory( sizeof( struct stack ));
   reclaimed_objects->values = memory( sizeof( void * ) * POOL_INC );
   reclaimed_objects->free = POOL_INC;
   reclaimed_objects->used = 0;
   reclaimed_objects->top = reclaimed_objects->values;

   reclaimed_atoms = ( struct stack *)memory( sizeof( struct stack ));
   reclaimed_atoms->values = memory( sizeof( void * ) * POOL_INC );
   reclaimed_atoms->free = POOL_INC;
   reclaimed_atoms->used = 0;
   reclaimed_atoms->top = reclaimed_atoms->values;

   object_pool_stack = make_stack();
   object_pool = ( struct object *)memory( sizeof( struct object ) * POOL_INC );
   object_pool_ptr = object_pool;
   object_pool_free = POOL_INC;

   atom_pool_stack = make_stack();
   atom_pool = ( struct tree *)memory( sizeof( struct tree ) * POOL_INC );
   atom_pool_free = POOL_INC;
   atom_pool_ptr = atom_pool;

   stack = make_stack();
   open_envs = make_stack();

   bookmark_stack = make_stack();
   buffer_stack = make_stack();
   seen = make_stack();
   token = make_string();

   tree = get_id( "lambda", 6 );
   tree->data.value = ( double *)1;
   tree->flags = ATOM_SYMBOL;
   set( tree->flags );

   lambda_id = tree->id;
   lambda_syntax = tree->syntax;

   tree = get_id( "macro", 5 );
   tree->data.value = ( double *)1;
   tree->flags = ATOM_SYMBOL;
   set( tree->flags );

   macro_id = tree->id;
   macro_syntax = tree->syntax;

   tree = get_id( "_", 1 );
   tree->data.value = ( double *)1;
   tree->flags = ATOM_SYMBOL;
   set( tree->flags );

   stop_id = tree->id;

   input_stack = make_stack();
   quotes_pending = make_stack();
   arg_stack = make_stack();

   init_intrinsics();

   zero = make_atom_from_number( 0 );
   empty = make_atom_from_string( "", 0 );

   pipe_pid[ 0 ] = -1;
   pipe_pid[ 1 ] = -1;
   pipe_pid[ 2 ] = -1;

   input_buffer_stack = make_stack();
   descriptors[ 0 ] = make_stack();
   descriptors[ 1 ] = make_stack();
   descriptors[ 2 ] = make_stack();

   result = 0;

   if ( stat( LISPDIR "/library.munger", &stats ) == 0 )
      result = load( LISPDIR "/library.munger" );
   else if ( stat( "./library.munger", &stats ) == 0 )
      result = load( "./library.munger" );
   else
      result = 1;

   if ( result )
   {
      fprintf( stderr, "munger: cannot load library.munger\n" );
      exit( 1 );
   }

   stack_pop( stack );

   ptr2 = getenv( "HOME" );

   if ( ptr2 != NULL )
   {
      s = make_string();
      string_assign( s, ptr2, strlen( ptr2 ));
      ptr2 = "/.munger";
      while( *ptr2 )
         string_append( s, *ptr2++ );

      if ( stat( s->str, &stats ) == 0 )
      {
         load( s->str );
         stack_pop( stack );
      }

      string_free( s );
   }

   if ( argc > 1 && load( argv[ 1 ] ) && fatal )
      exit( 1 );
}

void stack_free( struct stack *a )
{
   free( a->values );
   free( a );
}

void free_pool()
{
   void *ptr;

   free( object_pool );

   while(( ptr = stack_pop( object_pool_stack )) != NULL )
      free( ptr );

   stack_free( object_pool_stack );
}

void free_atoms()
{
   struct tree *ptr;

   free_syntax( syntax_tree );

   free( atom_pool );

   while(( ptr = ( struct tree *)stack_pop( atom_pool_stack )) != NULL )
      free( ptr );

   stack_free( atom_pool_stack );
}

void close_descriptors()
{
   int i;

   for( i = 0; i < 3; ++i )
      while( descriptors[ i ]->used )
         resume( "cleanup", i );
}

/*
 * This function is not strictly necessary, as all resources will
 * be released upon exit.
 */

void cleanup()
{
   int i;

   do_lisp_close( "cleanup", NULL );

   close_descriptors();

   while( input_buffer_stack->used )
   {
      stack_pop( input_buffer_stack );
      free( stack_pop( input_buffer_stack ));
   }

   stack_free( input_buffer_stack );

   for( i = 0; i < 3; ++i )
   {
      stack_free( descriptors[ i ] );

      /* 
       * Closes current std descriptors.  If redirected onto processes,
       * generates SIGCHLD, causes handler to reap zombies.
       */

      close( i );
   }

   hash_free( env );
   hash_free( atoms );

   free_atoms();
   free_pool();

   stack_free( stack );
   stack_free( open_envs );

   stack_free( input_stack );
   stack_free( quotes_pending );
   stack_free( arg_stack );
   stack_free( reclaimed_objects );
   stack_free( reclaimed_atoms );

   string_free( token );
   stack_free( seen );

   while( bookmark_stack->used )
   {
      if ( *bookmark_stack->top != NULL )
      {
         hash_free( *( struct hash_elt ***)bookmark_stack->top );
         free( *( struct hash_elt ***)bookmark_stack->top );
      }

      stack_pop( bookmark_stack );
   }

   stack_free( bookmark_stack );

   while( buffer_stack->used )
   {
      DB *db;

      db = ( DB *)stack_pop( buffer_stack );
      if ( db != NULL )
         db->close( db );
   }

   stack_free( buffer_stack );

   regfree( &find_poss_regex );
   regfree( &merge_regex );
}

void stack_push( struct stack *a, void *o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( void * ) * ( a->used + stack_inc ) );

      if ( a->values == NULL )
      {
         fprintf( stderr, "realloc: %s.\n", strerror( errno ));
         exit( 1 );
      }

      a->free = stack_inc;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   *a->top = o;
   --a->free;
   ++a->used;
}

struct object *stack_pop( struct stack *a )
{
   struct object *ptr;

   if ( a->used )
   {
      ptr = *a->top;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return ptr;
   }

   return NULL;
}

void stack_truncate( struct stack *a, int i )
{
   while( i-- )
      stack_pop( a );
}

void string_prepend( struct string *s, char c )
{
   char *ptr, *ptr2;

   if ( s->used == 0 )
   {
      string_append( s, c );
      return;
   }

   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + stack_inc );

      if ( s->str == NULL )
      {
         fprintf( stderr, "realloc: %s.\n", strerror( errno ));
         exit( 1 );
      }

      s->free = stack_inc;
      s->top = &s->str[ s->used ];
   }

   ptr2 = &s->str[ s->used + 1 ];

   for( ptr = &s->str[ s->used ]; ptr >= s->str; --ptr )
      *ptr2-- = *ptr;

   s->str[ 0 ] = c;

   ++s->used;
   ++s->top;
   --s->free;
}

void string_append( struct string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + stack_inc );

      if ( s->str == NULL )
      {
         fprintf( stderr, "realloc: %s.\n", strerror( errno ));
         exit( 1 );
      }

      s->free = stack_inc;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';
}

void string_erase( struct string *s, int idx )
{
   char *ptr, *ptr2;

   ptr = &s->str[ idx ];
   ptr2 = &s->str[ idx + 1 ];

   while( *ptr2 )
      *ptr++ = *ptr2++;

   *ptr = '\0';

   --s->used;
   ++s->free;
   --s->top;
}

void string_chop( struct string *s )
{
   if ( s->used )
   {
      *--s->top = '\0';
      --s->used;
      ++s->free;
   }
}

void string_truncate( struct string *s )
{
   if ( s->used )
   {
      s->free += s->used;
      s->used = 0;
      *s->str = '\0';
      s->top = s->str;
   }
}

void string_assign( struct string *s, char *c, int len )
{
   string_truncate( s );

   while( len-- )
      string_append( s, *c++ );
}

int get_token( int depth, int file )
{
   static char characters[ 102400 ] = "", stdin_buffer[ 102400 ] ="";
   static char *ptr = characters;
   static int count = 1;
   static int old_file = -1;
   static int escape = 0;
   int type;

   if ( file < 0 )
   {
      *characters = '\0';
      ptr = characters;
      old_file = -1;
      return -1;
   }

   type = -1;
   string_truncate( token );

   if ( old_file != file )
   {
      if ( old_file >= 0 )
      {
         if ( old_file == STDIN_FILENO )
            strcpy( stdin_buffer, ptr );
         else
         {
            int pos = -( strlen( ptr ));

            if ( lseek( old_file, pos, SEEK_CUR ) < 0 &&
                 errno != EBADF &&
                 errno != ESPIPE )
            {
               fprintf( stderr, "get_token: fatal error: lseek: %s.\n", 
                        strerror( errno ));
               exit( 1 );
            }
         }
      }

      *characters = '\0';
      ptr = characters;

      if ( file == STDIN_FILENO )
         ptr = stdin_buffer;
   }

   old_file = file;

   for( ; ; )
   {
      int i;

      if ( *ptr == '\0' )
      {
         if ( file == STDIN_FILENO )
         {
            if ( isatty( fileno( stdout )))
            {
               fputs( ">", stdout );

               i = depth;
               while( i-- > 0 )
                  fputs( ">", stdout );

               fputs( " ", stdout );
            }

            fflush( stdout );
         }

   AGAIN:
         if (( count = read( file, characters, sizeof( characters ) - 1 )) 
            == 0 )
         {
            if ( type >= 0 )
               return type;

            old_file = -1;
            return ( type = TOK_END );
         }
         else if ( count == -1 )
         {
            if ( errno == EINTR )
               goto AGAIN;

            fprintf( stderr, "read: %s.\n", strerror( errno ));
            exit( 1 );
         }

         characters[ count ] = '\0';
         ptr = characters;
      }

      for( ; *ptr; )
      {
         if ( type == TOK_COMMENT )
         {
            if ( *ptr == '\n' || *ptr == '\r' )
            {
               string_truncate( token );
               type = -1;
            }

            ++ptr;
            continue;
         }

         if ( *ptr == '\\' )
         {
            if ( type == TOK_STRING )
            {
               if ( !( escape ^= 1 ) )
                  string_chop( token );
            }
            else 
            {
               ++ptr;
               continue;
            }
         }
         else if ( *ptr != '"' )
            escape = 0;

         if ( *ptr == ';' || *ptr == '#' )
         {
            if ( type < 0 )
               type = TOK_COMMENT;
            else if ( type != TOK_STRING )
               return type;

            string_append( token, *ptr++ );
         }
         else if ( *ptr == '(' || *ptr == ')' )
         {
            if ( type >= 0 )
            {
               if ( type == TOK_STRING )
               {
                  string_append( token, *ptr++ );
                  continue;
               }

               return type;
            }

            type = ( *ptr == '(' ? TOK_OPEN : TOK_CLOSE );
            string_append( token, *ptr++ );
            return type;
         }
         else if ( *ptr == '"' )
         {
            if ( type < 0 )
            {
               type = TOK_STRING;
               string_append( token, *ptr++ );
            }
            else if ( type != TOK_STRING )
               return type;
            else
            {
               if ( escape )
                  string_chop( token );

               string_append( token, *ptr++ );

               if ( !escape )
                  return type;

               escape = 0;
            }
         }
         else if (( *ptr >= 'A' && *ptr <= 'Z' ) || 
                  ( *ptr >= 'a' && *ptr <= 'z' ) || *ptr == '_' )
         {
            if ( type == -1 )
               type = TOK_SYMBOL;
            else if ( type != TOK_STRING && type != TOK_SYMBOL )
               return type;

            string_append( token, *ptr++ );
         }
         else if ( *ptr == '.' || ( *ptr >= '0' && *ptr <= '9' ))
         {
            if ( type == -1 || type == TOK_MINUS )
               type = TOK_NUMBER;
            else if ( type != TOK_STRING && type != TOK_SYMBOL &&
                      type != TOK_NUMBER )
               return type;

            string_append( token, *ptr++ );
         }
         else if ( !isspace( *ptr ))
         {
            if ( type == -1 )
            {
               if ( *ptr == '\'' )
                  type = TOK_QUOTE;
               else if ( *ptr == '-' )
                  type = TOK_MINUS;
               else
                  type = TOK_SPECIAL;
            }
            else if ( type != TOK_STRING && 
                      type != TOK_SPECIAL )
               return type;

            string_append( token, *ptr++ );
         }
         else 
         {
            if ( type >= 0 )
            {
               if ( type != TOK_STRING )
                  return type;
               else if ( *ptr == '\r' )
               {
                  if ( isatty( STDIN_FILENO ))
                  {
                     *token->top = '\n';
                     ++ptr;
                  }
                  else
                     string_append( token, *ptr++ );
               }
               else 
                  string_append( token, *ptr++ );

               continue;
            }
            else
            {
               ++ptr;
               continue;
            }

            string_append( token, *ptr++ );
         }
      }
   }

   return type;
}

int process_token( int type, int depth )
{
   struct object *ptr;
   static int quote = -2;

   if ( type == TOK_OPEN )
   {
      if ( quote == -1 )
         quote = depth;

      ++depth;

      ptr = make_object();
      ptr->flags = -1;
      stack_push( input_stack, ptr );
   }
   else if ( type == TOK_QUOTE )
   {
      string_assign( token, "(", 1 );
      depth = process_token( TOK_OPEN, depth );

      string_assign( token, "quote", 5 );
      process_token( TOK_SYMBOL, depth );

      if ( quote > 0 )
         stack_push( quotes_pending, ( void *)quote );

      quote = -1;
   }
   else if ( type == TOK_CLOSE )
   {
      if ( --depth < 0 )
         return depth;

      if ( ( *( struct object **)input_stack->top )->flags == -1 )
         ( *( struct object **)input_stack->top )->flags = 1;
      else
      {
         for( ; ; )
         {
            ptr = stack_pop( input_stack );

            if (( *( struct object **)input_stack->top )->flags == -1 )
            {
               ( *( struct object **)input_stack->top )->flags = 1;
               ( *( struct object **)input_stack->top )->data.head = ptr;
               break;
            }
            else
               ( *( struct object **)input_stack->top )->next = ptr;
         }
      }

      if ( quote == depth )
      {
         if ( quotes_pending->used )
            quote = ( int)stack_pop( quotes_pending );
         else
            quote = -2;

         string_assign( token, ")", 1 );
         depth = process_token( TOK_CLOSE, depth );
      }
      else if ( depth == 0 )
         stack_push( stack, stack_pop( input_stack ));
   }
   else
   {
      struct tree *entry;
      double n = 0;

      if ( type == TOK_NUMBER )
      {
         char buf[ 64 ];

         n = atof( token->str );
         snprintf( buf, sizeof( buf ), "%.10g", n );

         entry = get_id( buf, strlen( buf ));
      }
      else
         entry = get_id( token->str, token->used );

      if ( entry->flags == 0 )
      {
         switch( type )
         {
            case TOK_STRING:
               token->str[ token->used - 1 ] = '\0';
               entry->flags = ATOM_STRING;
               entry->data.string = memory( sizeof( struct lstring ));
               entry->data.string->string = str_dup( &token->str[ 1 ], token->used - 2 );
               entry->data.string->length = token->used - 2;
               break;

            case TOK_SYMBOL:
            case TOK_SPECIAL:
               entry->flags = ATOM_SYMBOL;
               entry->data.value = NULL;
               break;

            case TOK_NUMBER:
               entry->flags = ATOM_NUMBER;
               entry->data.value = memory( sizeof( double ));
               *entry->data.value = n;
         }
      }

      ptr = make_object();
      ptr->data.atom = entry;

      stack_push( ( depth ? input_stack : stack ), ptr );

      if ( quote == -1 )
      {
         if ( quotes_pending->used )
            quote = ( int )stack_pop( quotes_pending );
         else
            quote = -2;

         string_assign( token, ")", 1 );
         depth = process_token( TOK_CLOSE, depth );
      }
   }

   return depth;
}

int parse( int file )
{
   int depth = 0;

   for( ; ; )
   {
      int type;

      type = get_token( depth, file );

      if ( type == TOK_END )
         return 1;

      depth = process_token( type, depth );

      if ( depth <= 0 )
         break;
   }

   return depth;
}

void do_printing( struct object *ptr, int recursive )
{
   while( ptr != NULL )
   {
      if ( islist( ptr->flags ) )
      {
         write( STDOUT_FILENO, "(", 1 );
         do_printing( ptr->data.head, 1 );
         write( STDOUT_FILENO, ")", 1 );
      }
      else
         write( STDOUT_FILENO, 
                ptr->data.atom->syntax,
                ptr->data.atom->len );

      if ( recursive == 0 )
         break;

      if (( ptr = ptr->next ) != NULL )
         write( STDOUT_FILENO, " ", 1 );
   }
}

struct object *make_atom_from_closure( struct closure *closure, int macro )
{
   struct tree *entry;
   struct object *obj;
   char buffer[ 128 ], *t;

   t = ( macro ? "<MACRO_CLOSURE#%d>" : "<CLOSURE#%d>" );

   snprintf( buffer, sizeof( buffer ), t, closure_counter++ );
   entry = get_id( buffer, strlen( buffer ));

   entry->flags = ( macro ? ATOM_MACRO : ATOM_CLOSURE );
   entry->data.closure = closure;

   obj = make_object();
   obj->data.atom = entry;

   return obj;
}

struct object *make_atom_from_act_record( struct stack *act_record )
{
   struct tree *entry;
   struct object *obj;
   char buffer[ 128 ];

   snprintf( buffer, sizeof( buffer ), "<ACT_RECORD#%d>", act_record_counter++ );
   entry = get_id( buffer, strlen( buffer ));

   entry->flags = ATOM_ACT_RECORD;
   entry->data.act_record = act_record;

   obj = make_object();
   obj->data.atom = entry;

   return obj;
}

int make_act_record( struct object *args, struct closure *closure, struct stack *act_record,
                     char *name, int eval_args  )
{
   int i, j, sym_count = 0, arg_count = 0;
   char *more = "";
   struct object *ptr, *ptr2;

   stack_push( act_record, closure->env );

   for( ptr = closure->text->data.head; ptr != NULL; ptr = ptr->next )
   {
      if ( ptr->next == NULL && islist( ptr->flags ))
         more = "(at least)";

      ++sym_count;
   }

   if ( sym_count && args == NULL && *more == '\0' )
   {
      fprintf( stderr, "%s: 0 argument(s) given to closure "
               "expecting %s %d argument(s).\n", name, more, sym_count );
      return 1;
   }

   for( ptr = args; ptr != NULL; ptr = ptr->next )
      ++arg_count;

   if ( sym_count != arg_count )
   {
      if ( *more == '\0' || ( arg_count < sym_count - 1 ))
      {
         fprintf( stderr, "%s: %d argument(s) given to closure"
                  " expecting %s %d argument(s).\n", 
                  name, arg_count, more, sym_count );
         return 1;
      }
   }

   if ( *more == '\0' )
   {
      if ( arg_count )
      {
         for( i = 1, ptr = closure->text->data.head, ptr2 = args;
              i <= arg_count;
              ptr = ptr->next, ptr2 = ptr2->next, ++i )
         {
            if ( eval_args )
            {
               stack_push( stack, ptr2 );

               if ( evaluate() )
               {
                  if ( !stop )
                     fprintf( stderr, "%s: evaluation of argument %d failed.\n",
                              name, i );
                  return 1;
               }
            }
            else
            {
               stack_push( act_record, ptr2 );
               stack_push( act_record, ( void *)ptr->data.atom->id );
            }
         }

         if ( eval_args )
         {
            for( j = i - 1, i = 1, ptr = closure->text->data.head; i <= arg_count; ptr = ptr->next, ++i, --j )
            {
               stack_push( act_record, stack->values[ stack->used - j ] );
               stack_push( act_record, ( void *)ptr->data.atom->id );
            }

            stack_truncate( stack, i - 1 );
         }
      }
   }
   else
   {
      struct object *result;

      if ( arg_count == 0 )
      {
         result = make_object();
         setlist( result->flags );
         stack_push( act_record, result );
         stack_push( act_record, ( void *)closure->text->data.head->data.head->data.atom->id );
      }
      else
      {
         struct object **ptr3;

         for( i = 1, ptr = closure->text->data.head, ptr2 = args;
              i < sym_count;
              ptr = ptr->next, ptr2 = ptr2->next, ++i )
         {
            if ( eval_args )
            {
               stack_push( stack, ptr2 );

               if ( evaluate() )
               {
                  if ( !stop )
                     fprintf( stderr, "%s: evaluation of argument %d failed.\n", name, i );
                  return 1;
               }
            }
            else
            {
               stack_push( act_record, ptr2 );
               stack_push( act_record, ( void *)ptr->data.atom->id );
            }
         }

         if ( i > 1 && eval_args )
         {
            for( j = i - 1, i = 1, ptr = closure->text->data.head; i < sym_count; ptr = ptr->next, ++i, --j )
            {
               stack_push( act_record, stack->values[ stack->used - j ] );
               stack_push( act_record, ( void *)ptr->data.atom->id );
            }

            stack_truncate( stack, i - 1 );
         }

         result = make_object();
         setlist( result->flags );
         stack_push( stack, result );
         stack_push( act_record, result );
         stack_push( act_record, ( void *)ptr->data.head->data.atom->id );

         ptr3 = &result->data.head;

         if ( ptr2 != NULL )
         {
            for( j = 1; ptr2 != NULL; ptr2 = ptr2->next, ++i, ++j )
            {
               if ( eval_args )
               {
                  stack_push( stack, ptr2 );

                  if ( evaluate() )
                  {
                     if ( !stop )
                        fprintf( stderr, "%s: evaluation of argument %d failed.\n", name, i );
                     return 1;
                  }
               }
               else
               {
                  *ptr3 = ptr2;
                  ptr3 = &( *ptr3 )->next;
               }
            }

            if ( eval_args )
            {
               ptr3 = &result->data.head;

               i = --j;

               for( ; j > 0; --j )
               {
                  *ptr3 = duplicate_object( stack->values[ stack->used - j ] );
                  ptr3 = &( *ptr3 )->next;
               }

               stack_truncate( stack, i );
            }
         }

         stack_pop( stack );

         *ptr3 = NULL;
      }
   }

   return 0;
}

int do_closure( char *name, struct closure *closure, struct object *args, int eval_args )
{
   struct object *obj;
   struct closure *old_closure;
   struct stack *act_record;
   int result = 0, i;

   act_record = make_stack();

   if ( make_act_record( args, closure, act_record, name, eval_args ))
      return 1;

   obj = make_atom_from_act_record( act_record );

   old_closure = current_closure;
   stack_push( open_envs, local_env );

   current_closure = closure;
   local_env = obj;

   i = stack->used;

TAILCALL:
   result = do_progn( name, closure->text->next );

   if ( result && tailcall )
   {
      stack_truncate( stack, stack->used - i );
      tailcall = stop = 0;
      goto TAILCALL;
   }

   current_closure = old_closure;
   local_env = stack_pop( open_envs );

   return result;
}

struct object *lookup_local( id )
{
   struct stack *local;
   struct object *env_ptr, **ptr;

   env_ptr = local_env;

   while( env_ptr != NULL )
   {
      local = env_ptr->data.atom->data.act_record;

      for( ptr = ( struct object **)local->top;
           ptr > ( struct object **)local->values;
           ptr -= 2 )
         if ( *( int *)ptr == id )
            return ( struct object *) *--ptr;

      env_ptr = ( struct object *)local->values[ 0 ];
   }

   return NULL;
}

int set_local( int id, struct object *obj )
{
   struct stack *local;
   struct object *env_ptr, **ptr;

   env_ptr = local_env;

   while( env_ptr != NULL )
   {
      local = env_ptr->data.atom->data.act_record;

      for( ptr = ( struct object **)local->top; 
           ptr > ( struct object **)local->values;
           ptr -= 2 )
         if ( *( int *)ptr == id )
         {
            *--ptr = obj;
            return 0;
         }

      env_ptr = ( struct object *)local->values[ 0 ];
   }

   return 1;
}

int make_closure( struct object *func, int macro )
{
   struct closure *closure;
   struct object *ptr;
   char *syntax = "make_closure";

   if ( islist( func->flags ) == 0 )
   {
      fprintf( stderr, "%s: missing parameter list.\n", syntax );
      return 1;
   }

   if ( func->next == NULL )
   {
      fprintf( stderr, "%s: missing function body.\n", syntax );
      return 1;
   }

   for( ptr = func->data.head; ptr != NULL; ptr = ptr->next )
   {
      if ( islist( ptr->flags ))
      {
         struct object *ptr2;

         if ( ptr->next != NULL )
         {
            fprintf( stderr, "%s: sublist in non-terminal position "
                             "of parameter list", syntax );
            return 1;
         }

         ptr2 = ptr->data.head;

         if ( islist( ptr2->flags ) || 
              type( ptr2->data.atom->flags ) != ATOM_SYMBOL )
         {
            fprintf( stderr, "%s: non-symbol in optional parameter sublist.\n", 
                     syntax );
            return 1;
         }

         if ( ptr2->next != NULL )
         {
            fprintf( stderr, "%s: optional parameter sublist has more than 1"
                             "element.\n", syntax );
            return 1;
         }
      }
      else if ( type( ptr->data.atom->flags ) != ATOM_SYMBOL )
      {
         fprintf( stderr, "%s: non-symbol in parameter list.\n", syntax );
         return 1;
      }
   }

   closure = memory( sizeof( struct closure ));
   closure->text = func;
   closure->env = local_env;

   stack_push( stack, make_atom_from_closure( closure, macro ));

   return 0;
}

int evaluate()
{
   static int gc = GC_FREQUENCY;
   int result = 0;
   struct object *item;

   if ( --gc == 0 )
   {
      collect_garbage();
      gc = GC_FREQUENCY;
   }

   item = stack_pop( stack );
   if ( item == NULL )
      return 1;

   if ( islist( item->flags ) == 0 )
   {
      struct object *value;
      int t;

      t = type( item->data.atom->flags );

      if ( item->data.atom->id == lambda_id ||
           item->data.atom->id == macro_id  ||
           t == ATOM_STRING    || 
           t == ATOM_NUMBER    ||
           t == ATOM_INTRINSIC ||
           t == ATOM_REGEXP    ||
           t == ATOM_DB        ||
           t == ATOM_STACK     ||
           t == ATOM_CLOSURE   ||
           t == ATOM_MACRO     ||
           t == ATOM_TABLE )
      {
         stack_push( stack, item );
         return 0;
      }

      if (( value = lookup_local( item->data.atom->id )) == NULL )
         value = lookup_binding( item->data.atom->id );

      if ( value == NULL )
      {
         fprintf( stderr, "evaluate: symbol %s not bound.\n", 
                  item->data.atom->syntax );
         return 1;
      }

      stack_push( stack, value );
   }
   else if ( item->data.head == NULL )
      stack_push( stack, item );
   else
   {
      char t, *name = NULL;
      struct object *car, *cdr;

      car = item->data.head;
      cdr = item->data.head->next;

      if ( islist( car->flags ) == 0 &&
           type( car->data.atom->flags ) == ATOM_SYMBOL )
         name = car->data.atom->syntax;

      stack_push( stack, item );
      stack_push( stack, car );

      if ( evaluate() )
         return 1;

      car = *( struct object **)stack->top;

      if ( islist( car->flags ))
      {
         fprintf( stderr, "evaluate: function position did not evaluate"
                          " to an intrinsic nor a closure.\n" );
         return 1;
      }

      if ( car->data.atom->id == lambda_id )
         result = make_closure( cdr, 0 );
      else if ( car->data.atom->id == macro_id )
         result = make_closure( cdr, 1 );
      else 
      {
#ifdef DEBUG
         int i;

         i = stack->used;
#endif

         t = type( car->data.atom->flags );

         switch( t )
         {
            case ATOM_INTRINSIC:
               result = car->data.atom->data.function( car->data.atom->syntax, cdr );
               break;

            case ATOM_CLOSURE:
               result = do_closure( ( name == NULL ? car->data.atom->syntax : name ),
                                    car->data.atom->data.closure,
                                    cdr,
                                    1 );
               break;

            case ATOM_MACRO:
               if (( result = do_closure( ( name == NULL ? car->data.atom->syntax : name ),
                                          car->data.atom->data.closure, 
                                          cdr,
                                          0 )) == 0 &&
                   ( result = evaluate() ))
               {
                  if ( !stop )
                     fprintf( stderr, "%s: secondary evaluation of macro failed.\n",
                              car->data.atom->syntax );
                  return 1;
               }
               break;

            default:
               fprintf( stderr, "evaluate: %s is not an intrinsic nor a closure.\n", 
                        car->data.atom->syntax );
               return 1;
         }

#ifdef DEBUG
         if ( i != stack->used - 1 && result != 1 )
         {
            fprintf( stderr, "stack corrupted by: %s: before: %d, after: %d\n", 
                     ( name == NULL ? car->data.atom->syntax : name ),
                     i, stack->used );
            exit( 1 );
         }
#endif
      }

      if ( result == 0 )
      {
         item = stack_pop( stack );
         stack_pop( stack );
         stack_pop( stack );
         stack_push( stack, item );
      }
   }

   return result;
}

void toplevel()
{
   char *t = "toplevel";
   int result;

   for( ; ; )
   {
      int depth;

      arg_stack->free += arg_stack->used;
      arg_stack->used = 0;
      arg_stack->top = arg_stack->values;

      stack->free += stack->used;
      stack->used = 0;
      stack->top = stack->values;

      if ( mode == 0 )
         canon( t );

      depth = parse( STDIN_FILENO );

      if ( depth > 0 )
         break;
      else if ( depth < 0 )
      {
         fprintf( stderr, "%s: %d extra ')'\n", t, -depth );

         if ( fatal )
            exit( 1 );
      }

      result = evaluate();

      close_descriptors();

      if ( result == 0 )
      {
         if ( printer )
         {
            print_object( *( struct object **)stack->top );
            putchar( '\n' );
         }
      }
      else if ( fatal )
         exit( 1 );

      stop = next_iteration = 0;

      if ( thrown != NULL )
      {
         fprintf( stderr, "%s: uncaught \"throw\"\n", t );

         if ( fatal )
            exit( 1 );

         thrown = NULL;
      }

      collect_garbage();
   }
}

void mark_object( struct object *obj )
{
   if ( obj == NULL )
      return;

   mark( obj->flags );

   if ( islist( obj->flags ) )
      mark_list( obj->data.head );
   else
   {
      set( obj->data.atom->flags );

      switch( type( obj->data.atom->flags ))
      {
         case ATOM_TABLE:
            mark_table( obj->data.atom->data.hash );
            break;

         case ATOM_STACK:
            mark_stack( obj->data.atom->data.stack );
            break;

         case ATOM_CLOSURE:
         case ATOM_MACRO:
            mark_closure( obj->data.atom->data.closure );
            break;

         case ATOM_ACT_RECORD:
            mark_act_record( obj->data.atom->data.act_record );
      }
   }
}

void mark_act_record( struct stack *act_record )
{
   int i;

   for( i = 0; i < seen->used; ++i )
      if ( seen->values[ i ] == act_record )
         return;

   stack_push( seen, act_record );

   mark_object( ( struct object *)act_record->values[ 0 ] );

   for( i = 1; i < act_record->used; i += 2 )
      mark_object( ( struct object *)act_record->values[ i ] );
}

void mark_closure( struct closure *closure )
{
   if ( closure == NULL )
      return;

   mark_list( closure->text );
   mark_object( closure->env );
}

void mark_stack( struct stack *stk )
{
   int i;

   if ( stk->used == 0 )
      return;

   for( i = 0; i < seen->used; ++i )
      if ( seen->values[ i ] == stk )
         return;

   stack_push( seen, stk );

   for( i = 0; i < stk->used; ++i )
      mark_object( ( struct object *)stk->values[ i ] );
}

void mark_table( struct hash_elt **ptr )
{
   struct stack *values;
   struct tree *key;
   int i;

   for( i = 0; i < seen->used; ++i )
      if ( seen->values[ i ] == ptr )
         return;

   stack_push( seen, ptr );

   values = get_hash_keys( ptr );

   while( values->used )
   {
      key = lookup_atom( ( int)stack_pop( values ));
      set( key->flags );
   }

   stack_free( values );

   values = get_hash_values( ptr );

   while( values->used )
      mark_object( ( struct object *)stack_pop( values ));

   stack_free( values );
}

void mark_list( struct object *ptr )
{
   while( ptr != NULL )
   {
      mark_object( ptr );
      ptr = ptr->next;
   }
}

void unmark_objects()
{
   struct object *obj;
   struct tree *atm;
   int i, j;

   i = object_pool_stack->used;

   while ( i )
   {
      obj = ( struct object *)object_pool_stack->values[ --i ];

      for( j = 0; j < POOL_INC; ++j, ++obj )
         unmark( obj->flags );
   }

   for( obj = object_pool; obj < object_pool_ptr; ++obj )
      unmark( obj->flags );

   i = atom_pool_stack->used;

   while( i )
   {
      atm = ( struct tree *)atom_pool_stack->values[ --i ];

      for( j = 0; j < POOL_INC; ++j, ++atm )
         if ( ! ( type( atm->flags ) == ATOM_INTRINSIC ||
                  ( type( atm->flags ) == ATOM_SYMBOL &&
                    atm->data.value == ( double *)1 )))
            unset( atm->flags );
   }

   for( atm = atom_pool; atm < atom_pool_ptr; ++atm )
   {
      if ( ! ( type( atm->flags ) == ATOM_INTRINSIC ||
               ( type( atm->flags ) == ATOM_SYMBOL &&
                 atm->data.value == ( double *)1 )))
         unset( atm->flags );
   }
}

void sweep_tree( struct tree *tree, struct tree **parent )
{
   if ( tree == NULL )
      return;

   if ( isitset( tree->flags ))
   {
      sweep_tree( tree->left, &tree->left );
      sweep_tree( tree->right, &tree->right );
      return;
   }

   free( tree->syntax );

   switch( type( tree->flags ))
   {
      case ATOM_STRING:
         free( tree->data.string->string );
         free( tree->data.string );
         break;

      case ATOM_REGEXP:
         regfree( tree->data.regexp );
         free( tree->data.regexp );
         break;

      case ATOM_TABLE:
         hash_free( tree->data.hash );
         free( tree->data.hash );
         break;

      case ATOM_DB:
#ifdef SQL
         if ( tree->data.db != NULL )
            sqlite3_close( tree->data.db );
#endif
         break;

      case ATOM_STACK:
         stack_free( tree->data.stack );
         break;

      case ATOM_NUMBER:
         free( tree->data.value );
         break;

      case ATOM_CLOSURE:
      case ATOM_MACRO:
         free( tree->data.closure );
         break;

      case ATOM_ACT_RECORD:
         stack_free( tree->data.act_record );
   }

   remove_elt( atoms, tree->id );
   stack_push( reclaimed_atoms, tree );

   tree->syntax = NULL;
   tree->len = 0;
   tree->id = 0;
   tree->flags = 0;
   tree->data.hash = NULL;

   if ( tree->left == NULL && tree->right == NULL )
   {
      *parent = NULL;
      return;
   }

   if ( tree->left == NULL )
      *parent = tree->right;
   else if ( tree->right == NULL )
      *parent = tree->left;
   else
   {
      int test;
      struct tree *ptr1, *ptr2;

      if (( random() * 2 ) >= 1 )
      {
         *parent = tree->right;
         ptr1 = tree->left;
      }
      else
      {
         *parent = tree->left;
         ptr1 = tree->right;
      }

      ptr2 = *parent;

   AGAIN:
      test = memcmp( ptr1->syntax, ptr2->syntax, MIN( ptr1->len, ptr2->len ) + 1 );

      if ( test > 0 )
      {
         if ( ptr2->right != NULL )
         {
            ptr2 = ptr2->right;
            goto AGAIN;
         }

         ptr2->right = ptr1;
      }
      else
      {
         if ( ptr2->left != NULL )
         {
            ptr2 = ptr2->left;
            goto AGAIN;
         }

         ptr2->left = ptr1;
      }
   }

   sweep_tree( *parent, parent );

   tree->left = tree->right = NULL;
}

void sweep_objects()
{
   struct object *top;
   int i, j, old_stack_inc;

   reclaimed_objects->free += reclaimed_objects->used;
   reclaimed_objects->used = 0;
   reclaimed_objects->top = reclaimed_objects->values;

   old_stack_inc = stack_inc;
   stack_inc = POOL_INC;

   i = object_pool_stack->used;

   while( i )
   {
      top = ( struct object *)object_pool_stack->values[ --i ];

      for( j = 0; j < POOL_INC; ++j, ++top )
         if ( ismarked( top->flags ) == 0 )
         {
            bzero( top, sizeof( struct object ));
            stack_push( reclaimed_objects, top );
         }
   }

   for( top = object_pool; top < object_pool_ptr; ++top )
   {
      if ( ismarked( top->flags ) == 0 )
      {
         bzero( top, sizeof( struct object ));
         stack_push( reclaimed_objects, top );
      }
   }

   stack_inc = old_stack_inc;
}

void mark_bookmarks()
{
   int i;
   struct stack *keys;
   struct tree *key;

   for( i = 0; i < bookmark_stack->used; ++i )
   {
      if ( bookmark_stack->values[ i ] == NULL )
         continue;

      keys = get_hash_keys( ( struct hash_elt **)bookmark_stack->values[ i ] );

      while( keys->used )
      {
         key = lookup_atom( ( int)stack_pop( keys ));
         set( key->flags );
      }

      stack_free( keys );
   }
}

void collect_garbage()
{
   seen->free += seen->used;
   seen->used = 0;
   seen->top = seen->values;

   mark_bookmarks();

   mark_object( zero );
   mark_object( empty );

   mark_object( local_env );
   mark_stack( open_envs );

   mark_table( env );
   mark_stack( stack );

   sweep_objects();
   sweep_tree( syntax_tree, &syntax_tree );
   unmark_objects();
}

void reset_term()
{
   if ( mode )
      canon( "" );
}

int main( int argc, char **argv )
{
   atexit( reset_term );

   initialize( argc, argv );
   toplevel();
   cleanup();

   return 0;
}
