/*
 * Copyright (c) 2001, 2002, 2003, 2004 James Bailie <jimmy@jamesbailie.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote 
 * products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <term.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <regex.h>
#include <dirent.h>
#include <termios.h>
#include <db.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <signal.h>
#include <assert.h>
#include <math.h>
#include <pwd.h>
#include <grp.h>
#include <curses.h>
#include <libgen.h>

#ifdef SQL
#include "sqlite3.h"
#endif

#ifdef FCGI
#include "fcgi_config.h"
#include "fcgiapp.h"
#endif

#define HASH_SIZE 415
#define POOL_INC 8192
#define GC_FREQUENCY 1000

#define print_object(p) do_printing(p,0)

#define FUNC(n) int do_ ## n ( char *, struct object * )

#define type(n) ( n & 127 )
#define set(n) ( n |= 128 )
#define unset(n) ( n &= 127 )
#define isitset(n) ( n & 128 )

#define mark(n) ( n |= 2 )
#define unmark(n) ( n &= 1 )
#define ismarked(n) ( n & 2 )

#define setlist(n) ( n |= 1 )
#define unlist(n) ( n &= 2 )
#define islist(n) ( n & 1 )

#define ATOM_SYMBOL 1
#define ATOM_STRING 2
#define ATOM_NUMBER 3
#define ATOM_REGEXP 4
#define ATOM_TABLE  5
#define ATOM_STACK  6
#define ATOM_INTRINSIC 7
#define ATOM_DB 8
#define ATOM_CLOSURE 9
#define ATOM_MACRO 10
#define ATOM_ACT_RECORD 11

enum token_types { TOK_STRING, TOK_SYMBOL, TOK_NUMBER, TOK_SPECIAL, TOK_OPEN,
                   TOK_CLOSE, TOK_END, TOK_QUOTE, TOK_COMMENT, TOK_MINUS };

enum err_types { ERR_SYMBOL,       ERR_STRING,      ERR_NUMBER,    
                 ERR_REGEXP,       ERR_TABLE,       ERR_STACK,       
                 ERR_DB,           ERR_INTRINSIC,   ERR_ATOM,
                 ERR_LIST,         ERR_CLOSURE,     ERR_ARGS,
                 ERR_MISSING_ARGS, ERR_MISSING_ARG, ERR_MORE_ARGS,
                 ERR_EVAL,         ERR_ARG_TYPE,    ERR_CONSTANT   };

char *err_strings[] = { "symbol", "string", "number", "regexp", "table", 
                        "stack", "database", "intrinsic", "atom", "list",
                        "closure" };

char *error_messages[] =
{ "%s: called with argument(s).\n",
  "%s: called without argument(s).\n",
  "%s: called without argument %d.\n",
  "%s: called with more than %d argument(s).\n",
  "%s: evaluation of argument %d failed.\n",
  "%s: argument %d did not evaluate to a %s.\n",
  "%s: argument %d evaluated to a constant.\n"
};

struct termios canon_termios;

char **first_arg, **last_arg, **arg_ptr;

void sigchld_handler( int );
void sigwinch_handler( int );
int pipe_open( char *, char *, int, int );
void print_err( enum err_types, char *, int, enum err_types );

struct stack
{
   int free, used;
   void **top, **values;
};

struct hash_elt
{
   struct object *element;
   struct hash_elt *next;
   int binding;
};

struct tree
{
   char *syntax;
   int len;
   int id;

   union {
      double *value;
      regex_t *regexp;
      struct hash_elt **hash;
      struct stack *stack;
#ifdef SQL
      sqlite3 *db;
#endif
      int ( *function )( char *, struct object * );
      struct closure *closure;
      struct stack *act_record;
      struct lstring { char *string; int length; } *string;
   } data;

   char flags;
   struct tree *left, *right;
};

struct object
{
   char flags;

   union {
      struct tree *atom;
      struct object *head;
   } data;

   struct object *next;
};

struct closure
{
   struct object *text, *env;
};

struct string
{
   int free, used;
   char *top;
   char *str;
};

DB *buffer = NULL;
DBT dbt_key, dbt_value;
recno_t key_data;

regex_t find_poss_regex, merge_regex;

int stack_inc = 16;
int mode = 1, printer = 1;

struct object *sql_list = NULL;

#ifdef FCGI
FCGX_Stream *cgi_in = NULL, *cgi_out = NULL, *cgi_err = NULL;
FCGX_ParamArray cgi_param;
#endif

struct hash_elt *env[ HASH_SIZE ];
struct hash_elt *atoms[ HASH_SIZE ];

struct object *local_env = NULL;
struct closure *current_closure = NULL;

struct stack *get_hash_keys( struct hash_elt ** );
struct stack *get_hash_values( struct hash_elt ** );
struct object *copy_object( struct object * );

void canon( char * );

struct tree *syntax_tree;
struct object *thrown = NULL, *zero, *empty;
struct string *token;

struct stack *input_stack, *quotes_pending, *arg_stack, *seen, *buffer_stack,
   *stack, *open_envs;

struct object *object_pool, *object_pool_ptr;
struct stack *object_pool_stack, *atom_pool_stack, *reclaimed_objects,
   *reclaimed_atoms;

int make_act_record( struct object *,
                     struct closure *,
                     struct stack *,
                     char *,
                     int );

struct tree *atom_pool, *old_atom_pool, *atom_pool_ptr;

int object_pool_free, atom_pool_free, fatal = 0, child_fd = -1,
   lambda_id, macro_id, stop_id, stop = 0, stack_counter = 1,
   closure_counter = 1, lisp_fd = -1, lisp_pid = -1, 
   tailcall = 0, act_record_counter = 1, db_counter = 1,
   table_counter = 1, rx_counter = 1, gensym_counter = 1, atom_counter = 1, 
   interactive = 0, next_iteration = 0, blocked = 0;

int pipe_pid[ 3 ];

struct stack *descriptors[ 3 ], *input_buffer_stack;

char *lambda_syntax, *macro_syntax;

struct hash_elt **bookmarks;
struct stack *bookmark_stack;
void adjust_bookmarks( int, int );
void delete_bookmarks( int, int );

struct object *make_object();
struct tree *get_id( char *, int len );

void *memory( int );

void unmark_objects();
void mark_bookmarks();
void mark_list( struct object * );
void mark_table( struct hash_elt ** );
void mark_stack( struct stack * );
void mark_closure( struct closure * );
void mark_act_record( struct stack * );
void mark_object( struct object * );
void sweep_tree( struct tree *, struct tree ** );
void collect_garbage();

int set_local( int, struct object * );

int load( char * );
void cleanup();
int evaluate();
void resume( char *, int );
int make_closure( struct object *, int );
int do_closure( char *, struct closure *, struct object *, int );
int get_token( int, int );
int parse( int );
int getline_from_file( char *, int );

struct stack *make_stack();
void stack_push( struct stack *, void * );
struct object *stack_pop( struct stack * );
void stack_truncate( struct stack *, int );
void stack_free( struct stack * );

void do_printing( struct object *ptr, int recursive );

void insert_elt( struct hash_elt **, int, struct object* );
void remove_elt( struct hash_elt **, int );
void *lookup_elt( struct hash_elt **, int );
void hash_free( struct hash_elt ** );

struct object *lookup_local( int id );

#define insert_binding(n,m) insert_elt( env, n, m )
#define insert_atom(n,m) insert_elt( atoms, n, ( struct object *)m )
#define lookup_atom(n) ( struct tree *)lookup_elt( atoms, n )
#define lookup_binding(n) ( struct object *)lookup_elt( env, n )
#define remove_binding(n) remove_elt( env, n )

struct object *make_atom_from_act_record( struct stack * );
struct object *make_atom_from_number( double );
struct object *make_atom_from_closure( struct closure *, int );

struct string *make_string();
void string_free( struct string * );
void string_prepend( struct string *, char );
void string_append( struct string *, char );
void string_chop( struct string * );
void string_truncate( struct string * );
void string_assign( struct string *, char *, int );
void string_erase( struct string *, int );

char *str_dup( char *, int );

FUNC( progn );
FUNC( cons );
FUNC( quote );
FUNC( car );
FUNC( cdr );
FUNC( eq );
FUNC( atom );
FUNC( append );
FUNC( set );
FUNC( eval );
FUNC( if );
FUNC( and );
FUNC( or );
FUNC( list );
FUNC( length );
FUNC( not );
FUNC( print );
FUNC( newline );
FUNC( load );
FUNC( getline );
FUNC( complete );
FUNC( cols );
FUNC( lines );
FUNC( numberp );
FUNC( stringp );
FUNC( split );
FUNC( join );
FUNC( match );
FUNC( matches );
FUNC( substitute );
FUNC( regcomp );
FUNC( sort );
FUNC( sortlist );
FUNC( sortcar );
FUNC( while );
FUNC( until );
FUNC( do );
FUNC( throw );
FUNC( catch );
FUNC( die );
FUNC( stringify );
FUNC( digitize );
FUNC( intern );

FUNC( additive );
FUNC( subtractive );
FUNC( comparative );

FUNC( abs );
FUNC( char );
FUNC( code );
FUNC( open );
FUNC( close );
FUNC( insert );
FUNC( delete );
FUNC( lastline );
FUNC( retrieve );
FUNC( filter );
FUNC( write );
FUNC( read );
FUNC( empty );
FUNC( slice );
FUNC( find );
FUNC( input );
FUNC( output );
FUNC( system );
FUNC( maxidx );
FUNC( chdir );
FUNC( mkdir );
FUNC( bound );
FUNC( buffer );
FUNC( buffers );
FUNC( switch );
FUNC( version );
FUNC( gensym );
FUNC( libdir );
FUNC( substring );
FUNC( expand );
FUNC( interact );
FUNC( current );
FUNC( next );
FUNC( prev );
FUNC( rewind );
FUNC( pwd );
FUNC( exit );
FUNC( table );
FUNC( hash );
FUNC( unhash );
FUNC( lookup );
FUNC( keys );
FUNC( values );
FUNC( redirect );
FUNC( resume );
FUNC( warn );
FUNC( getenv );
FUNC( directory );
FUNC( chomp );
FUNC( chop );
FUNC( unlink );
FUNC( rmdir );
FUNC( rename );
FUNC( words );
FUNC( date );
FUNC( unless );
FUNC( when );
FUNC( test );
FUNC( continue );
FUNC( block );
FUNC( unblock );
FUNC( exists );
FUNC( suspend);
FUNC( beep );
FUNC( int );
FUNC( tan );
FUNC( atan );
FUNC( cos );
FUNC( acos );
FUNC( sin );
FUNC( asin );
FUNC( sqrt );
FUNC( cbrt );
FUNC( strcmp );
FUNC( fatal );
FUNC( nofatal );
FUNC( let );
FUNC( letn );
FUNC( labels );
FUNC( cond );

#ifdef SQL
FUNC( sql_open );
FUNC( sql_close );
FUNC( sql_exec );
FUNC( dbp );
#endif

#ifdef FCGI
FUNC( cgi_accept );
FUNC( cgi_finish );
FUNC( cgi_xstatus );
FUNC( cgi_read );
FUNC( cgi_print );
FUNC( cgi_newline );
FUNC( cgi_getenv );
FUNC( cgi_warn );
FUNC( cgi_sendfile );
#endif

FUNC( format );
FUNC( stack );
FUNC( clear );
FUNC( push );
FUNC( pop );
FUNC( shift );
FUNC( unshift );
FUNC( used );
FUNC( store );
FUNC( index );
FUNC( time );
FUNC( random );
FUNC( topidx );
FUNC( extract );
FUNC( transfer );
FUNC( pipe );
FUNC( setenv );

FUNC( lisp_open );
FUNC( lisp_running );
FUNC( lisp_read );
FUNC( lisp_write );
FUNC( lisp_close );
FUNC( lisp_ready );

FUNC( protect );
FUNC( tailcall );
FUNC( stat );
FUNC( realpath );
FUNC( access );
FUNC( power );

FUNC( setmark );
FUNC( getmark );

FUNC( noprinter );
FUNC( printer );
FUNC( exec );
FUNC( truncate );

FUNC( symbolp );
FUNC( regexpp );
FUNC( tablep );
FUNC( stackp );
FUNC( intrinsicp );
FUNC( closurep );
FUNC( macrop );

FUNC( dynamic_let );
FUNC( chmod );
FUNC( chown );
FUNC( basename );
FUNC( dirname );
FUNC( checkpass );
FUNC( setuid );
FUNC( seteuid );
FUNC( getuid );
FUNC( geteuid );
FUNC( getgid );

FUNC( seek );
FUNC( getchars );
FUNC( readlock );
FUNC( writelock );
FUNC( unlock );
FUNC( hostname );
FUNC( symlink );
FUNC( gecos );

#ifdef TERM
FUNC( pause );
FUNC( display );
FUNC( scrollup );
FUNC( scrolldn );
FUNC( clearscreen );
FUNC( goto );
FUNC( clearline );
FUNC( hide );
FUNC( show );
FUNC( insertln );
FUNC( getchar );
FUNC( pushback );
FUNC( canon );
FUNC( nocanon );

char *cl = ( char *)-1, *ce = ( char *)-1, *cm = ( char *)-1, 
   *ti = ( char *)-1, *te = ( char *)-1,
   *vi = ( char *)-1, *ve = ( char *)-1, *sf = ( char *)-1,
   *sr = ( char *)-1, *al = ( char *)-1, *sc = ( char *)-1,
   *rc = ( char *)-1;

int pushed_back = -1, sigwinch = 0;
#endif

int ( *intrinsics[] )( char *, struct object * ) =
{
   do_progn, 
   do_cons,
   do_quote,
   do_car,
   do_cdr,
   do_eq,
   do_atom,
   do_append,
   do_set,
   do_eval,
   do_if,
   do_and,
   do_or,
   do_list,
   do_not,
   do_print,
   do_newline,
   do_load,
   do_getline,
   do_complete,
   do_cols,
   do_lines,
   do_numberp,
   do_stringp,
   do_split,
   do_join,
   do_match,
   do_matches,
   do_substitute,
   do_regcomp,
   do_sort,
   do_sortlist,
   do_sortcar,
   do_while,
   do_until,
   do_do,
   do_throw,
   do_catch,
   do_die,
   do_stringify,
   do_digitize,
   do_intern,

/*
 * These intrinsics determine which operation to perform from their
 * syntax arguments.
 */

   do_additive,
   do_subtractive,
   do_additive,
   do_subtractive,
   do_subtractive,
   do_comparative,
   do_comparative,
   do_comparative,
   do_comparative,

   do_abs,
   do_char,
   do_code,
   do_open,
   do_close,
   do_insert,
   do_delete,
   do_lastline,
   do_retrieve,
   do_filter,
   do_write,
   do_read,
   do_empty,
   do_slice,
   do_find,
   do_input,
   do_output,
   do_system,
   do_maxidx,
   do_chdir,
   do_mkdir,
   do_bound,
   do_buffer,
   do_buffers,
   do_switch,
   do_version,
   do_gensym,
   do_libdir,
   do_substring,
   do_expand,
   do_interact,
   do_current,
   do_next,
   do_prev,
   do_rewind,
   do_pwd,
   do_exit,
   do_table,
   do_hash,
   do_unhash,
   do_lookup,
   do_keys,
   do_values,
   do_redirect,
   do_resume,
   do_warn,
   do_getenv,
   do_directory,
   do_chomp,
   do_chop,
   do_unlink,
   do_rmdir,
   do_rename,
   do_words,
   do_date,
   do_when,
   do_unless,
   do_test,
   do_continue,
   do_block,
   do_unblock,
   do_exists,
   do_suspend,
   do_beep,
   do_length,
   do_int,
   do_tan,
   do_atan,
   do_cos,
   do_acos,
   do_sin,
   do_asin,
   do_sqrt,
   do_cbrt,
   do_strcmp,
   do_fatal,
   do_nofatal,
   do_let,
   do_letn,
   do_labels,
   do_cond,

#ifdef SQL
   do_sql_open,
   do_sql_close,
   do_sql_exec,
   do_dbp,
#endif

#ifdef FCGI
   do_cgi_accept,
   do_cgi_finish,
   do_cgi_xstatus,
   do_cgi_read,
   do_cgi_print,
   do_cgi_newline,
   do_cgi_getenv,
   do_cgi_warn,
   do_cgi_sendfile,
#endif

   do_format,
   do_stack,
   do_clear,
   do_push,
   do_pop,
   do_used,
   do_store,
   do_index,
   do_time,
   do_random,
   do_topidx,
   do_extract,
   do_transfer,
   do_pipe,
   do_setenv,
   do_lisp_open,
   do_lisp_running,
   do_lisp_read,
   do_lisp_write,
   do_lisp_close,
   do_lisp_ready,
   do_protect,
   do_tailcall,
   do_stat,
   do_realpath,
   do_access,
   do_power,
   do_setmark,
   do_getmark,
   do_shift,
   do_unshift,

#ifdef TERM
   do_pause,
   do_display,
   do_scrollup,
   do_scrolldn,
   do_clearscreen,
   do_goto,
   do_clearline,
   do_hide,
   do_show,
   do_insertln,
   do_getchar,
   do_pushback,
   do_canon,
   do_nocanon,
#endif

   do_noprinter,
   do_printer,
   do_exec,
   do_truncate,

   do_symbolp,
   do_regexpp,
   do_tablep,
   do_stackp,
   do_intrinsicp,
   do_closurep,
   do_macrop,
   do_dynamic_let,
   do_chmod,
   do_chown,
   do_basename,
   do_dirname,
   do_checkpass,
   do_setuid,
   do_seteuid,
   do_getuid,
   do_getgid,
   do_geteuid,
   do_seek,
   do_getchars,
   do_readlock,
   do_writelock,
   do_unlock,
   do_hostname,
   do_symlink,
   do_gecos,
   NULL
};

char *intrinsic_syntax[] =
{
   "<PROGN>",         "progn",
   "<CONS>",          "cons",
   "<QUOTE>",         "quote",
   "<CAR>",           "car",
   "<CDR>",           "cdr",
   "<EQ>",            "eq",
   "<ATOM>",          "atom",
   "<APPEND>",        "append",
   "<SET>",           "set",
   "<EVAL>",          "eval",
   "<IF>",            "if",
   "<AND>",           "and",
   "<OR>",            "or",
   "<LIST>",          "list",
   "<NOT>",           "not",
   "<PRINT>",         "print",
   "<NEWLINE>",       "newline",
   "<LOAD>",          "load",
   "<GETLINE>",       "getline",
   "<COMPLETE>",      "complete",
   "<COLS>",          "cols",
   "<LINES>",         "lines",
   "<NUMBERP>",       "numberp",
   "<STRINGP>",       "stringp",
   "<SPLIT>",         "split",
   "<JOIN>",          "join",
   "<MATCH>",         "match",
   "<MATCHES>",       "matches",
   "<SUBSTITUTE>",    "substitute",
   "<REGCOMP>",       "regcomp",
   "<SORT>",          "sort",
   "<SORTLIST>",      "sortlist",
   "<SORTCAR>",       "sortcar",
   "<WHILE>",         "while",
   "<UNTIL>",         "until",
   "<DO>",            "do",
   "<THROW>",         "throw",
   "<CATCH>",         "catch",
   "<DIE>",           "die",
   "<STRINGIFY>",     "stringify",
   "<DIGITIZE>",      "digitize",
   "<INTERN>",        "intern",
   "<PLUS>",          "+",
   "<MINUS>",         "-",
   "<TIMES>",         "*",
   "<DIVIDE>",        "/",
   "<MODULO>",        "%",
   "<LESSER>",        "<",
   "<LESSER_OR_EQ>",  "<=",
   "<GREATER>",       ">",
   "<GREATER_OR_EQ>", ">=",
   "<ABS>",           "abs",
   "<CHAR>",          "char",
   "<CODE>",          "code",
   "<OPEN>",          "open",
   "<CLOSE>",         "close",
   "<INSERT>",        "insert",
   "<DELETE>",        "delete",
   "<LASTLINE>",      "lastline",
   "<RETRIEVE>",      "retrieve",
   "<FILTER>",        "filter",
   "<WRITE>",         "write",
   "<READ>",          "read",
   "<EMPTY>",         "empty",
   "<SLICE>",         "slice",
   "<FIND>",          "find",
   "<INPUT>",         "input",
   "<OUTPUT>",        "output",
   "<SYSTEM>",        "system",
   "<MAXIDX>",        "maxidx",
   "<CHDIR>",         "chdir",
   "<MKDIR>",         "mkdir",
   "<BOUND>",         "bound",
   "<BUFFER>",        "buffer",
   "<BUFFERS>",       "buffers",
   "<SWITCH>",        "switch",
   "<VERSION>",       "version",
   "<GENSYM>",        "gensym",
   "<LIBDIR>",        "libdir",
   "<SUBSTRING>",     "substring",
   "<EXPAND>",        "expand",
   "<INTERACT>",      "interact",
   "<CURRENT>",       "current",
   "<NEXT>",          "next",
   "<PREV>",          "prev",
   "<REWIND>",        "rewind",
   "<PWD>",           "pwd",
   "<EXIT>",          "exit",
   "<TABLE>",         "table",
   "<HASH>",          "hash",
   "<UNHASH>",        "unhash",
   "<LOOKUP>",        "lookup",
   "<KEYS>",          "keys",
   "<VALUES>",        "values",
   "<REDIRECT>",      "redirect",
   "<RESUME>",        "resume",
   "<WARN>",          "warn",
   "<GETENV>",        "getenv",
   "<DIRECTORY>",     "directory",
   "<CHOMP>",         "chomp",
   "<CHOP>",          "chop",
   "<UNLINK>",        "unlink",
   "<RMDIR>",         "rmdir",
   "<RENAME>",        "rename",
   "<WORDS>",         "words",
   "<DATE>",          "date",
   "<WHEN>",          "when",
   "<UNLESS>",        "unless",
   "<TEST>",          "test",
   "<CONTINUE>",      "continue",
   "<BLOCK>",         "block",
   "<UNBLOCK>",       "unblock",
   "<EXISTS>",        "exists",
   "<SUSPEND>",       "suspend",
   "<BEEP>",          "beep",
   "<LENGTH>",        "length",
   "<INT>",           "int",
   "<TAN>",           "tan",
   "<ATAN>",          "atan",
   "<COS>",           "cos",
   "<ACOS>",          "acos",
   "<SIN>",           "sin",
   "<ASIN>",          "asin",
   "<SQRT>",          "sqrt",
   "<CBRT>",          "cbrt",
   "<STRCMP>",        "strcmp",
   "<FATAL>",         "fatal",
   "<NOFATAL>",       "nofatal",
   "<LET>",           "let",
   "<LETN>",          "letn",
   "<LABELS>",        "labels",
   "<COND>",          "cond",

#ifdef SQL
   "<SQL_OPEN>",      "sql_open",
   "<SQL_CLOSE>",     "sql_close",
   "<SQL_EXEC>",      "sql_exec",
   "<DBP>",           "dbp",
#endif

#ifdef FCGI
   "<CGI_ACCEPT>",    "cgi_accept",
   "<CGI_FINISH>",    "cgi_finish",
   "<CGI_XSTATUS>",   "cgi_xstatus",
   "<CGI_GETLINE>",   "cgi_read",
   "<CGI_PRINT>",     "cgi_print",
   "<CGI_NEWLINE>",   "cgi_newline",
   "<CGI_GETENV>",    "cgi_getenv",
   "<CGI_WARN>",      "cgi_warn",
   "<CGI_SENDFILE>",  "cgi_sendfile",
#endif

   "<FORMAT>",        "format",
   "<STACK>",         "stack",
   "<CLEAR>",         "clear",
   "<PUSH>",          "push",
   "<POP>",           "pop",
   "<USED>",          "used",
   "<STORE>",         "store",
   "<INDEX>",         "index",
   "<TIME>",          "time",
   "<RANDOM>",        "random",
   "<TOPIDX>",        "topidx",
   "<EXTRACT>",       "extract",
   "<TRANSFER>",      "transfer",
   "<PIPE>",          "pipe",
   "<SETENV>",        "setenv",
   "<LISP_OPEN>",     "lisp_open",
   "<LISP_RUNNING>",  "lisp_running",
   "<LISP_READ>",     "lisp_read",
   "<LISP_WRITE>",    "lisp_write",
   "<LISP_CLOSE>",    "lisp_close",
   "<LISP_READY>",    "lisp_ready",
   "<PROTECT>",       "protect",
   "<TAILCALL>",      "tailcall",
   "<STAT>",          "stat",
   "<REALPATH>",      "realpath",
   "<ACCESS>",        "access",
   "<POWER>",         "power",
   "<SETMARK>",       "setmark",
   "<GETMARK>",       "getmark",
   "<SHIFT>",         "shift",
   "<UNSHIFT>",       "unshift",

#ifdef TERM
   "<PAUSE>",         "pause",
   "<DISPLAY>",       "display",
   "<SCROLLUP>",      "scrollup",
   "<SCROLLDN>",      "scrolldn",
   "<CLEARSCREEN>",   "clearscreen",
   "<GOTO>",          "goto",
   "<CLEARLINE>",     "clearline",
   "<HIDE>",          "hide",
   "<SHOW>",          "show",
   "<INSERTLN>",      "insertln",
   "<GETCHAR>",       "getchar",
   "<PUSHBACK>",      "pushback",
   "<CANON>",         "canon",
   "<NOCANON>",       "nocanon",
#endif

   "<NOPRINTER>",     "noprinter",
   "<PRINTER>",       "printer",
   "<EXEC>",          "exec",
   "<TRUNCATE>",      "truncate",

   "<SYMBOLP>",       "symbolp",
   "<REGEXPP>",       "regexpp",
   "<TABLEP>",        "tablep",
   "<STACKP>",        "stackp",
   "<INTRINSICP>",    "intrinsicp",
   "<CLOSUREP>",      "closurep",
   "<MACROP>",        "macrop",

   "<DYNAMIC_LET>",   "dynamic_let",
   "<CHMOD>",         "chmod",
   "<CHOWN>",         "chown",
   "<BASENAME>",      "basename",
   "<DIRNAME>",       "dirname",
   "<CHECKPASS>",     "checkpass",
   "<SETUID>",        "setuid",
   "<SETEUID>",       "seteuid",
   "<GETUID>",        "getuid",
   "<GETGID>",        "getgid",
   "<GETEUID>",       "geteuid",
   "<SEEK>",          "seek",
   "<GETCHARS>",      "getchars",
   "<READLOCK>",      "readlock",
   "<WRITELOCK>",     "writelock",
   "<UNLOCK>",        "unlock",
   "<HOSTNAME>",      "hostname",
   "<SYMLINK>",       "symlink",
   "<GECOS>",         "gecos"
};
