/* 
 * Copyright (c) 2004 James Bailie <jimmy@jamesbailie.com>.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote 
 * products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This filter converts an XML document into an alist, similar to the one
 * which would be produced by xml2lisp.munger, with two differences.
 * First, an alist representing an empty element cannot be differentiated
 * from an non-empty element with no content.  That is to say
 * <element></element> and <element/> will both be represented by 
 * (element () ()).  Second, cdata content may be spread across more
 * than one cdata element, even if the parent element contains no embedded
 * child elements to break up the cdata.  See the comments in xml2lisp.munger
 * and xmlquery.munger for more information on using the alist produced by
 * this filter.
 */ 

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "expat.h"

void start( void *data, const char *element, const char **attrs )
{
   printf( "(\"%s\" (", element );

   if ( *attrs )
   {
      char **ptr;

      ptr = ( char **)attrs;

      while( *ptr )
      {
         char *ptr2;

         printf( "(\"" );

         ptr2 = *ptr++;

         while( *ptr2 )
         {
            if ( *ptr2 == '"' || *ptr2 == '\\' )
               putchar( '\\' );

            putchar( *ptr2++ );
         }

         printf( "\" \"" );

         ptr2 = *ptr++;

         while( *ptr2 )
         {
            if ( *ptr2 == '"' || *ptr2 == '\\' )
               putchar( '\\' );

            putchar( *ptr2++ );
         }

         printf( "\")" );
      }
   }

   printf( ") (" );
}

void end( void *data, const char *element )
{
   printf( "))" );
}

void cdata( void *data, const XML_Char *s, int len )
{
   char *ptr;

   printf( "(cdata \"" );

   ptr = ( char *)s;

   while( len-- )
   {
      if ( *ptr == '"' || *ptr == '\\' )
         putchar( '\\' );

      putchar( *ptr++ );
   }

   printf( "\")" );
}

int main( int argc, char **argv )
{
   char buffer[ 1024000 ];
   FILE *file;
   int result;
   XML_Parser p;

   p = XML_ParserCreate( NULL );

   if ( p == NULL )
   {
      fprintf( stderr, "XML_ParserCreate failed.\n" );
      exit( 1 );
   }

   XML_SetElementHandler( p, start, end );
   XML_SetCharacterDataHandler( p, cdata );

   if ( argc > 1 )
   {
      file = fopen( argv[ 1 ], "r" );

      if ( file == NULL )
      {
         fprintf( stderr, "fopen: %s.\n", strerror( errno ));
         exit( 1 );
      }
   }
   else
      file = stdin;

   printf( "(setq document '((\"document\" () (" );

   while(( result = fread( buffer, 1, sizeof( buffer ), file )))
   {
      if ( XML_Parse( p, buffer, result, 0 ) == 0 )
      {
         fprintf( stderr, "XML_parse: line %d: %s\n",
                  XML_GetCurrentLineNumber( p ),
                  XML_ErrorString( XML_GetErrorCode( p )));
         exit( 1 );
      }
   }

   printf( "))))" );

   if ( ferror( file ))
   {
      fprintf( stderr, "fread: %s\n", strerror( errno ));
      exit( 1 );
   }
   else if ( XML_Parse( p, "", 0, 1 ) == 0 )
   {
      fprintf( stderr, "XML_parse: line %d: %s\n",
               XML_GetCurrentLineNumber( p ),
               XML_ErrorString( XML_GetErrorCode( p )));
      exit( 1 );
   }

   XML_ParserFree( p );

   return 0;
}
